/* 
    fastburden - generates a flood of web traffic

  This is a performance testing tool that can generate a test flood of
  client web traffic towards a web server. It reads from the file "sites"
  and will sequentially fetch the URLs in a loop. This can be used 
  to assist in performance tuning of web and web proxy servers.

  Custom URLs can be added to the sites file to refelect realistic traffic
  for your web server, popular sites (eg /index.html) can be repeated in the
  file so that they are tested more often.

  The sites file could be generated by taking the correct field from the
  web server or proxy server log (eg Apache, Squid, ...) - which would
  execute a replay of actual client traffic. This replays the traffic as
  fast as possible (depending on the number of threads) ignoring the time 
  delays between site hits. (see squid2sites.sh and apache2sites.pl, 
  contained in this distribution).

  This is a companion to fastbrute, the mock FTP brute force attack tool.

 USAGE:
	fastburden [-p port] [-t threads] [-f file] [-r rounds]
		   [-d delay ms ]  hostname

  -p destination port, default is 80. Use other ports (eg 3128, 8080) to
     run against other types of servers.

  -t number of threads, default is 8 (max 250). By increasing or decreasing 
     the number of threads the speed of the attack can be controlled.
     Time a few trials with "-r 10" to see the effect.

  -f input file of URLs, default is "sites". 
  
  -r numbers of rounds of the sites file to loop through, default is 65536.

  -d extra ms delay between fetches (use this if the program is running 
     too fast; eg, "-d 100").

 EXAMPLES:
	./fastburden -p 8080 mars	# testing squid
	./fastburden -t 32 mars		# 32 threads 
	time ./fastburden -r 2 mars	# time 2 rounds of the sites list

 INSTALLING:
	make
	make install
  Edit the Makefile if the build fails and use settings appropriate for
  your OS.

 VERSION: 07-Nov-2004, ver 0.61

 WARNING: Please don't use this software for anything illegal. That 
  definition differs for every country, please check the law first.
  This is a performace testing tool - not a Denial of Service 
  or "hacking" tool - a misidentification that could render owning
  this software illegal in some countries.

 COPYRIGHT: Copyright (c) 2003 Brendan Gregg.

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version. 

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details. 

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software Foundation, 
  Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  (http://www.gnu.org/copyleft/gpl.html)

 AUTHOR: Brendan Gregg  [Sydney, Australia].

 TODO:
  * Test against different web servers.

 HISTORY:
	01-Dec-2003	Brendan Gregg	Created this.
	02-Dec-2003	   "      "	Optimised threading strategy.
*/

#include <unistd.h>
#include <stdlib.h>
#include <pthread.h>
#include <stdio.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h> 
#include <strings.h>
#include <string.h>


/*********************
    Global constants
**********************/

/* Default max number of threads running at once */
#define DEFAULT_THREADS 8

/* Max number of threads possible */
#define MAX_THREADS 250

/* Max number of sites that can be loaded */
#define MAX_SITES 8192

/* print debug messages */
#define DEBUG 0

/* print tried site count messages */
#define TRIED_DEBUG 1

/* max wait seconds for threads to complete */
#define MAX_SECONDS_WAIT 8

/* max extra wait seconds per thread */
#define MAX_SECONDS_EXTRA 1

/* filenames */
char *file_sites = "sites";	/* Sites input filename */


/*********************
    Global variables
**********************/

/* mutex locks for the threads */
pthread_mutex_t mutex_trys;
pthread_mutex_t mutex_running;

/* sites lists */
char *Sites[MAX_SITES];

/* number of currently running threads */
int threads_running;

/* 
**  Try struct
**
**  This is used as the main inter-thread communication method
**  (main thread <-> worker threads). It is mutex locked each time
**  main or a worker thread writes to it (not for all reads).
*/
typedef struct Try {
	int done[MAX_THREADS];	/* status of thread, 0 = done, any other
				   number is used as timeout "chalk" marks */
	int port;		/* dest port */
	int total;		/* total number of attempted sites */
	int delay;		/* ms delay between attempts (default 0) */
	int site;		/* current site number from list */
	int site_max;		/* max site in list */
	int round;		/* current round (loop) of the list */
	int round_max;		/* max rounds to run through */
	char *hostname;		/* dest hostname */
	char *sites[MAX_SITES];	/* sites list */
} Try;

/* try info holder */
Try try;


/*************************
    Subroutine Predecs
**************************/

void *thread_function(void *arg);	/* worker thread */
void die(char *message, int status);	/* error message and exit */
int load_sites();			/* load sites list */



/*******************
*   MAIN PROGRAM   *
********************/

/* main - The main program.
**	the main thread keeps an eye on the global int "mutex_running",
**	which is decremented each time a thread completes - reaching zero
**	when all threads have completed.
*/
int main(int argc, char *argv[]) {
   /*
   **  Variable Declaration
   */
   int count;				/* counter */
   int site_max;			/* max element of Sites array */
   int site;				/* site currently trying */
   int wait_seconds;			/* thread timeout seconds */
   int num_threads;			/* max num threads to run */
   int last_round;			/* last round of threads? */
   int total;				/* total sites tried so far */
   int option;				/* command line switch */
   int round;				/* round number (round of sites) */
   int round_max;			/* max number of rounds */
   int port;				/* port to connect to */
   int done;				/* thread status */
   int delay;				/* delay between fetches */
   pthread_t p_thread[MAX_THREADS];	/* POSIX thread objects */
   char *hostname;			/* target hostname */
   char *usage;				/* USAGE message */
   struct hostent *test_addr;		/* only for testing in main */

   /* defaults */
   round = 1; round_max = 65536;
   last_round = 0; total = 0; site = 0; delay = 0;
   num_threads = DEFAULT_THREADS;
   port = 80;
   usage = "USAGE: fastburden [-p port] [-t threads] [-f file] [-r rounds]\n\t\t  [-d delay ms]  hostname";

   fprintf(stdout,"fastburden ver 0.61\n");


   /*
   **  Command line arguments
   */

   /* check arg count */
   if (argc < 2) die(usage,0);

   /* process switches */
   while ((option = getopt(argc,argv,"d:f:p:r:t:")) != -1) {
	switch (option) {
		case 'p':
			port = atoi((void *) optarg);
			break;
		case 'r':
			round_max = atoi((void *) optarg);
			break;
		case 'd':
			delay = atoi((void *) optarg);
			break;
		case 't':
			num_threads = atoi((void *) optarg);
			if (num_threads > MAX_THREADS) 
				num_threads = MAX_THREADS;
			if (num_threads < 1) num_threads = 1;
			break;
		case 'f':
			file_sites = strdup(optarg);
			break;
		default:
			die(usage,1);
	}
   }
   argv += optind;
   if (! *argv) die(usage,1);
   hostname = *argv;


   /* calculate timeout seconds for every round of threads */
   wait_seconds = MAX_SECONDS_WAIT + (MAX_SECONDS_EXTRA * num_threads);

   /*** get target hostname ***/
   test_addr = gethostbyname(hostname);
   if (test_addr == NULL) {
   	die("ERROR: No such host",1);
   }

   /*
   ** Load Sites
   */
   site_max = load_sites();
   if (DEBUG) fprintf(stderr,"00. smax %d port %d threads %d file %s\n",
    site_max,port,num_threads,file_sites);
   fprintf(stdout,"Started...  (Loaded %d sites)\n",(site_max+1));

   /*
   **  Populate Work Plan
   */
   try.port = port;
   try.hostname = strdup(hostname);
   for(count = 0; count <= site_max; count++) {
	try.sites[count] = Sites[count];
   }
   for(count = 0; count <= num_threads; count++) {
	try.done[count] = 0;
   }
   try.round = round;
   try.round_max = round_max;
   try.site_max = site_max;
   try.total = 0;
   try.delay = delay;

   /* initialise mutex locks, just in case */
   pthread_mutex_init(&mutex_running,NULL);
   pthread_mutex_init(&mutex_trys,NULL);

   /* decreases during final round only */
   threads_running = num_threads;

	
   /*************
      Main Loop
   **************/

   while (1) {

	/*
	**  Prepare and Create the threads
	*/
	for (count = 0; count < num_threads; count++) {
		if (DEBUG) fprintf(stderr,"00. getting %d tr %d\n",
		 site,threads_running);

		/*
		**  Get thread status
		*/
		pthread_mutex_lock(&mutex_trys);
		done = try.done[count];
		if (!done) try.done[count]++;
		pthread_mutex_unlock(&mutex_trys);

		/*
		**  Timeout stuck threads
		*/
		if ((done * 5) > wait_seconds) {
			pthread_cancel(p_thread[count]);
			pthread_join(p_thread[count], NULL);
			pthread_mutex_lock(&mutex_trys);
			try.done[count] = 0;
			pthread_mutex_unlock(&mutex_trys);
			done = 0;
		}

		if (try.round > try.round_max) {
			last_round = 1;		/* as it's the last round */
		} 

		/*
		**  Thread Creation
		*/
		if ((!(last_round))&&(done == 0)) {
			/*
			**  Create the thread
			*/
			if (pthread_create(&p_thread[count],NULL, 
				thread_function,(void *) count) != 0) {
					fprintf(stderr,
					 "ERROR8: creating a thread\n");
			} 
		} 
		if ((last_round)&&(done == 0)) {
			/* this one wasn't needed */
			pthread_mutex_lock(&mutex_running);
			threads_running--;
			pthread_mutex_unlock(&mutex_running);
		}

		total = try.total; /* ro - no lock */

		if (threads_running == 0) 
			break;

	}
	if (threads_running == 0) 
		break;

	/*** wait a little. tunable ***/
	usleep(200000); /* 1/5 sec */

	if (TRIED_DEBUG) fprintf(stdout,"sent %d\n",total);
   }

   /*
   **  Cleanup Threads
   */

   /*** If some threads were still stuck, cancel them ***/
   for (count = 0; count < num_threads; count++) {
	if (try.done[count] == 0) {
		pthread_cancel(p_thread[count]);
		pthread_join(p_thread[count], NULL);
	}
   }

   if (DEBUG) fprintf(stderr, "00. All threads completed.\n");

   if (last_round) {
	if (TRIED_DEBUG) fprintf(stdout,"sent %d\n",total);
	fprintf(stdout,"Finished.\n");
	exit(0);
   }

   return 0;
}



/******************
*   SUBROUTINES   *
*******************/

/* die - this subroutine prints an error message and exits with a non-zero
** 	exit status.
*/
void die(char *message, int status) {
	fprintf(stderr,"%s\n",message);
	exit(status);
}

/* load_sites - this subroutine loads the sites from the file in the
**	global "file_sites" into the global Sites. 
*/
int load_sites() {
	char buffer[256];		/* file read buffer */
	char copy[256];			/* for string manipulation */
	FILE *infile;			/* filehandle */
	int site_max;			/* max element of Site array */
	int length;			/* length for strings */

	/*
	** Load Sites
	*/
	site_max = 0;
	if (!( infile = fopen(file_sites,"r") )) {
		die("ERROR9: Can't open sites file",9);
	}
	while ( fgets (buffer,256,infile) ) {
		length = strlen(buffer);
		bzero(copy,256);

		/* strip off the "\n" */
		strncpy(copy,buffer,(length-1));

		/* store sites in memory */
		Sites[site_max] = strdup(copy);
		if (DEBUG) fprintf (stderr,"site %d %s",site_max,buffer);

		/* count how many sites were loaded */
		site_max++;

		/* quit before you seg fault */
		if (site_max == MAX_SITES) break;
	}
	fclose(infile);
	
	/* that last site wasn't loaded */
	site_max--;

	return site_max;
}


/* thread_functions - The worker thread.
**      This subroutine connects to the host and fetches a site.
**	This is invoked numerous times (num_threads) as a worker thread. 
**	It is possible that the connection could hang for some reason
**	the main thread detects this by watching "mutex_running" and 
**	can send a pthread_cancel.
*/
void *thread_function(void *arg) {
	/*
	**  Variable Declaration
	*/
        struct sockaddr_in destination;	/* dest socket address */
        struct hostent *dest_addr;	/* dest address */
        char buffer[256];		/* file buffer */
	char *site;			/* site to fetch */
	char *hostname;			/* dest hostname */
	char send_str[256];		/* string to send */
        int sock;			/* socket number */
	int length;			/* length for strings */
	int dest_port;			/* dest port number */
	int id;				/* thread id */
	int num;			/* current try num */
	int max;			/* try max */
	int delay;			/* ms delay */

	/* thread number */
	id = (int) arg;

	if (DEBUG) fprintf(stderr, "%2d: Thread Started.\n", id);

	/* 
	**  Fetch host info
	*/
	hostname = strdup(try.hostname); /* ro - no lock */
	dest_port = try.port;
	max = try.site_max;
	delay = try.delay;

	/* hostname and port */
	dest_addr = gethostbyname(hostname);
	if (dest_addr == NULL) {
               	die("ERROR1: No such host\n",1);
	}

	/*
	**  Create destination address variable
	*/
	bzero((char *) &destination, sizeof(destination));
	destination.sin_family = AF_INET;
	bcopy((char *)dest_addr->h_addr, 
        ( char *)&destination.sin_addr.s_addr,
         dest_addr->h_length);
	destination.sin_port = htons(dest_port);


	while (1) {
		/* 
		**  Fetch and increment site number.
		**  Mutex locks are used here as many threads could be
		**  accessing the same variable.
		*/
		pthread_mutex_lock(&mutex_trys);
		num = try.site;
		try.site++;
		pthread_mutex_unlock(&mutex_trys);
		if (num > max) num = 0; 
		
		/* record total count */
		pthread_mutex_lock(&mutex_trys);
		try.total++;
		pthread_mutex_unlock(&mutex_trys);
	
		/*
		**  Fetch try site
		*/
		site = strdup(try.sites[num]); /* ro - no lock */

		/*
		**  Socket Creation
		*/
		sock = socket(AF_INET, SOCK_STREAM, 0);
		if (sock < 0) die("ERROR2: opening socket",2);

		/*
		**  Connect to destination
		*/
		if (connect(sock,(struct sockaddr *)&destination,
		 sizeof(destination)) < 0) 
			die("ERROR3: connecting",3); 

		/*
		**  send site
		*/
		sprintf(send_str,"GET %s HTTP/1.0\n\n",site);
		strcpy(buffer,send_str);
		if (DEBUG) fprintf(stderr,"%2d: TRYING %s \n",id,site);
		length = write(sock,buffer,strlen(buffer));
		if (length < 0) die("ERROR5: writing to socket",5);

		/*
		**  read reply
		*/
		bzero(buffer,256);
		length = read(sock,buffer,255);
		if (length < 0) die("ERROR6: reading from socket",6);
		if (DEBUG > 1) printf("%2d: %d.1 %s",id,num,buffer);

		close(sock);
	
		/*
		**  Let main know I am OK
		*/
		pthread_mutex_lock(&mutex_trys);
		try.done[id] = 1;
		pthread_mutex_unlock(&mutex_trys);

		/* delay if required */
		if (delay) usleep (delay * 1000);

		/* check for end point */
		if (try.site > max) {
			pthread_mutex_lock(&mutex_trys);
			try.round++;
			try.site = 0;
			pthread_mutex_unlock(&mutex_trys);
		}
		if (try.round > try.round_max) {
			break;
		} 
	}

	if (DEBUG) fprintf(stderr, "%2d. Thread Finished.\n", id);

	/*** if main hasn't cancelled me, exit nicely ***/
	pthread_mutex_lock(&mutex_trys);
	try.done[id] = 0;
	pthread_mutex_unlock(&mutex_trys);
	pthread_exit(0);

	return 0;
}


/*********************
*   END OF PROGRAM   *
**********************/


/***************************************************************************/

/* Reminders for myself ... */

/* ----------------- */
/* Indentation style */

/* tabs (tabstop 8). triple space only occasionally. margin of 80 chars. 
   code that has line wrapped is indented with one space. eg,
					fprintf(stdout,"the number is %d\n",
					 number);	 */


/* --------------- */
/* Shortcuts style */

/* the inclination here is not to use them. eg,

	if (num++ > 6)
		num = 0;

   would more likely be written as,

	if (num > 6) {
		num = 0;
	}
	num++;

   for clarity. especially if the code is more complex than the above. */


/* --------------- */
/* Comments style, */

/* Micro comment */

/*** Tiny comment ***/

/*
**  Small Comment
*/

/********************
     Medium Comment 
*********************/

/************************
*     LARGE COMMENT     *
*************************/
